from __future__ import annotations

import subprocess

from cleo.exceptions import CommandNotFoundException
from cleo.helpers import argument
from cleo.io.outputs.output import Verbosity
from poetry.console.commands.command import Command

from poetry_aliases_plugin import config, utils
from poetry_aliases_plugin.aliases import AliasesSet
from poetry_aliases_plugin.config import AliasesConfig
from poetry_aliases_plugin.triggers import TriggerCommand


class BaseAliasCommand(Command):
    @property
    def aliases_config(self):
        return AliasesConfig(self.poetry.pyproject)

    def __find_poetry_command(self, command: str) -> tuple[str, str]:
        if not self.aliases_config.settings['find_poetry_command']:
            return 'run', command

        args = command.split()

        poetry_command = args[0]

        try:
            self.application.get(poetry_command)

        except CommandNotFoundException:
            poetry_command = 'run'

        else:
            args = args[1:]

        return poetry_command, ' '.join(args)

    def __poetry_call(self, poetry_command: str, args: str):
        self.call(poetry_command, args)

    def __subprocess_call(self, poetry_command: str, string_args: str):
        args = string_args.split()
        args = ['poetry', poetry_command] + args

        return subprocess.run(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        )

    def __exec_command(self, poetry_command: str, args: str):
        if self.aliases_config.settings['run_with_poetry']:
            self.__poetry_call(poetry_command, args)

        if self.aliases_config.settings['run_with_subprocess']:
            self.__subprocess_call(poetry_command, args)

    def exec_command(self, command: str):
        self.io.write(f'run command: {command}', True, Verbosity.DEBUG)

        poetry_command, args = self.__find_poetry_command(command)

        try:
            self.__exec_command(poetry_command, args)

        except PermissionError as ex:
            if ex.errno == 13:
                raise utils.PluginCommandException(command, f'У процесса poetry недостаточно прав для запуска программы: {ex.filename}')

            raise utils.PluginCommandException(command, ex.args[0])

        raise utils.PluginCommandException(command, 'Так быть не должно...')

    @property
    def trigger_command(self):
        raise NotImplementedError()

    @utils.plugin_exception_wrapper
    def handle(self) -> None:
        self.io.write(f'Trigger command: {self.trigger_command}', True, Verbosity.DEBUG)

        self.aliases_config.validate()

        aliases_set = AliasesSet.from_raw(self.aliases_config.aliases)
        triggered_aliases = aliases_set.get_triggered_aliases(self.trigger_command)

        for alias in triggered_aliases:
            self.io.write(f'run alias: {alias.alias}', True, Verbosity.DEBUG)

            for command in alias.commands:
                self.exec_command(command)


class AliasCommand(BaseAliasCommand):
    name = 'l'

    arguments = [argument('alias', 'Registered alias')]

    @property
    def description(self):
        return 'Run aliases. Available: {0}'.format(', '.join(list(self.aliases_config.aliases)))

    @property
    def trigger_command(self):
        return TriggerCommand(self.argument('alias'))
